import pytest
import numpy as np
import tonic.transforms as transforms
import itertools
from utils import create_random_input


@pytest.mark.parametrize(
    "sensor_size, size",
    itertools.product(((120, 120, 2), (120, 30, 2)), (50, (50, 50))),
)
def test_center_crop(sensor_size, size):
    orig_events, _ = create_random_input(sensor_size=sensor_size)

    transform = transforms.CenterCrop(sensor_size=sensor_size, size=size)

    events = transform(orig_events)

    if type(size) == int:
        size = [size, size]
    assert all(0 <= events["x"]) and all(events["x"] < size[0])
    assert all(0 <= events["y"]) and all(events["y"] < size[1])


@pytest.mark.parametrize("min, max", itertools.product((0, 1000), (None, 5000)))
def test_crop_time(min, max):
    orig_events, sensor_size = create_random_input()

    transform = transforms.CropTime(min=min, max=max)

    events = transform(orig_events)

    assert events is not orig_events
    if min is not None:
        assert not events["t"][0] < min
    if max is not None:
        assert not events["t"][-1] > max


@pytest.mark.parametrize("filter_time", [10000, 5000])
def test_transform_denoise(filter_time):
    orig_events, sensor_size = create_random_input()

    transform = transforms.Denoise(filter_time=filter_time)

    events = transform(orig_events)

    assert len(events) > 0, "Not all events should be filtered"
    assert len(events) < len(
        orig_events
    ), "Result should be fewer events than original event stream"
    assert np.isin(events, orig_events).all(), (
        "Denoising should not add additional events that were not present in"
        " original event stream"
    )
    assert events is not orig_events


@pytest.mark.parametrize(
    "p",
    [
        0.2,
        0.5,
        (0.1, 0.6),
    ],
)
def test_transform_drop_events(p):
    orig_events, sensor_size = create_random_input()

    if type(p) == tuple:
        sampled_p = transforms.DropEvent.get_params(p=p)
        assert p[0] <= sampled_p <= p[1]
        events = transforms.functional.drop_event_numpy(
            events=orig_events, drop_probability=sampled_p
        )
        p = sampled_p

    else:
        transform = transforms.DropEvent(p=p)
        events = transform(orig_events)

    assert np.isclose(events.shape[0], round((1 - p) * orig_events.shape[0])), (
        "Event dropout should result in p*len(original) events" " dropped out."
    )
    assert np.isclose(
        np.sum((events["t"] - np.sort(events["t"])) ** 2), 0
    ), "Event dropout should maintain temporal order."
    assert events is not orig_events


@pytest.mark.parametrize(
    "duration_ratio",
    [(0.1), (0.2), (0.3), (0.4), (0.5), (0.6), (0.7), (0.8), (0.9)],
)
def test_transform_drop_events_by_time(duration_ratio):
    orig_events, sensor_size = create_random_input()

    transform = transforms.DropEventByTime(duration_ratio=duration_ratio)

    events = transform(orig_events)

    assert len(events) < len(orig_events)

    t_start = orig_events["t"].min()
    t_end = orig_events["t"].max()

    # checks that there is no events during a period of the defined duration ratio
    duration = (t_end - t_start) * duration_ratio

    diffs = np.diff(events["t"])

    assert np.any(
        diffs >= duration
    ), f"There should be no events during {duration} in the obtained sequence."


@pytest.mark.parametrize(
    "area_ratio",
    [(0.1), (0.2), (0.3), (0.4), (0.5), (0.6), (0.7), (0.8), (0.9)],
)
def test_transform_drop_events_by_area(area_ratio):
    orig_events, sensor_size = create_random_input()

    transform = transforms.DropEventByArea(sensor_size, area_ratio)

    events = transform(orig_events)

    assert len(events) < len(orig_events)  # events were dropped by the transform

    # checks that an area of the right dimension contains no events in the resulting sequence
    cut_w = int(area_ratio * sensor_size[0])
    cut_h = int(area_ratio * sensor_size[1])

    to_im = transforms.ToImage(sensor_size)
    frame = to_im(events)
    orig_frame = to_im(orig_events)
    cmp = frame - orig_frame
    dropped_events = len(orig_events) - len(events)

    # goal: find the area that contains the same number of dropped events
    dropped_area_found = False
    for bbx1 in range(0, (sensor_size[0] - cut_w)):
        bbx2 = bbx1 + cut_w
        for bby1 in range(0, (sensor_size[1] - cut_h)):
            bby2 = bby1 + cut_h

            if abs(np.sum(cmp[:, bby1:bby2, bbx1:bbx2])) == dropped_events:
                dropped_area_found = True
                break

    assert (
        dropped_area_found is True
    ), f"There should be an area with {dropped_events} events dropped in the obtained sequence."


def test_transform_decimation():
    n = 10

    orig_events, sensor_size = create_random_input(sensor_size=(1, 1, 2), n_events=1000)
    transform = transforms.Decimation(n=n)

    events = transform(orig_events)

    assert len(events) == 100


@pytest.mark.parametrize(
    "coordinates, hot_pixel_frequency",
    [(((9, 11), (10, 12), (11, 13)), None), (None, 10000)],
)
def test_transform_drop_pixel(coordinates, hot_pixel_frequency):
    orig_events, sensor_size = create_random_input(sensor_size=(20, 20, 2))
    orig_events = np.concatenate((orig_events, np.ones(10000, dtype=orig_events.dtype)))
    orig_events = orig_events[np.argsort(orig_events["t"])]

    transform = transforms.DropPixel(
        coordinates=coordinates, hot_pixel_frequency=hot_pixel_frequency
    )

    events = transform(orig_events)

    assert len(events) < len(orig_events)

    if coordinates:
        for x, y in coordinates:
            assert not np.logical_and(events["x"] == x, events["y"] == y).sum()

    if hot_pixel_frequency:
        assert not np.logical_and(events["x"] == 1, events["y"] == 1).sum()
    assert events is not orig_events


@pytest.mark.parametrize(
    "coordinates, hot_pixel_frequency",
    [(((9, 11), (10, 12), (11, 13)), None), (None, 5000)],
)
def test_transform_drop_pixel_raster(coordinates, hot_pixel_frequency):
    raster_test = np.random.randint(0, 100, (50, 2, 100, 200))
    frame_test = np.random.randint(0, 100, (2, 100, 200))
    transform = transforms.DropPixel(
        coordinates=coordinates, hot_pixel_frequency=hot_pixel_frequency
    )
    raster = transform(raster_test)
    frame = transform(frame_test)

    if coordinates:
        for x, y in coordinates:
            assert raster[:, :, x, y].sum() == 0
            assert frame[:, x, y].sum() == 0
    if hot_pixel_frequency:
        merged_polarity_raster = raster.sum(0).sum(0)
        merged_polarity_frame = frame.sum(0)
        assert not merged_polarity_frame[merged_polarity_frame > 5000].sum().sum()
        assert not merged_polarity_raster[merged_polarity_raster > 5000].sum().sum()


@pytest.mark.parametrize("time_factor, spatial_factor", [(1, 0.25), (1e-3, 1)])
def test_transform_downsample(time_factor, spatial_factor):
    orig_events, sensor_size = create_random_input()

    transform = transforms.Downsample(
        time_factor=time_factor, spatial_factor=spatial_factor
    )

    events = transform(orig_events)

    assert np.array_equal(
        (orig_events["t"] * time_factor).astype(orig_events["t"].dtype), events["t"]
    )
    assert np.array_equal(np.floor(orig_events["x"] * spatial_factor), events["x"])
    assert np.array_equal(np.floor(orig_events["y"] * spatial_factor), events["y"])
    assert events is not orig_events


@pytest.mark.parametrize("target_size", [(50, 50), (10, 5)])
def test_transform_random_crop(target_size):
    orig_events, sensor_size = create_random_input()

    transform = transforms.RandomCrop(sensor_size=sensor_size, target_size=target_size)
    events = transform(orig_events)

    assert np.all(events["x"]) < target_size[0] and np.all(
        events["y"] < target_size[1]
    ), "Cropping needs to map the events into the new space."
    assert events is not orig_events


@pytest.mark.parametrize("p", [1.0, 0])
def test_transform_flip_lr(p):
    orig_events, sensor_size = create_random_input()

    transform = transforms.RandomFlipLR(sensor_size=sensor_size, p=p)

    events = transform(orig_events)

    if p == 1:
        assert ((sensor_size[0] - 1) - orig_events["x"] == events["x"]).all(), (
            "When flipping left and right x must map to the opposite pixel, i.e. x' ="
            " sensor width - x"
        )
    else:
        assert np.array_equal(orig_events, events)
    assert events is not orig_events


@pytest.mark.parametrize("p", [1.0, 0])
def test_transform_flip_polarity(p):
    orig_events, sensor_size = create_random_input()

    transform = transforms.RandomFlipPolarity(p=p)

    events = transform(orig_events)

    if p == 1:
        assert np.array_equal(np.invert(orig_events["p"].astype(bool)), events["p"]), (
            "When flipping polarity with probability 1, all event polarities must"
            " flip"
        )
    else:
        assert np.array_equal(orig_events["p"], events["p"]), (
            "When flipping polarity with probability 0, no event polarities must"
            " flip"
        )
    assert events is not orig_events


@pytest.mark.parametrize("p", [1.0, 0])
def test_transform_flip_polarity_bools(p):
    orig_events, sensor_size = create_random_input(
        dtype=np.dtype([("x", int), ("y", int), ("t", int), ("p", bool)])
    )

    transform = transforms.RandomFlipPolarity(p=p)

    events = transform(orig_events)

    if p == 1:
        assert np.array_equal(np.invert(orig_events["p"].astype(bool)), events["p"]), (
            "When flipping polarity with probability 1, all event polarities must"
            " flip"
        )
    else:
        assert np.array_equal(orig_events["p"], events["p"]), (
            "When flipping polarity with probability 0, no event polarities must"
            " flip"
        )
    assert events is not orig_events


@pytest.mark.parametrize("p", [1.0, 0])
def test_transform_flip_ud(p):
    orig_events, sensor_size = create_random_input()

    transform = transforms.RandomFlipUD(sensor_size=sensor_size, p=p)

    events = transform(orig_events)

    if p == 1:
        assert np.array_equal((sensor_size[1] - 1) - orig_events["y"], events["y"]), (
            "When flipping left and right x must map to the opposite pixel, i.e. x' ="
            " sensor width - x"
        )
    else:
        assert np.array_equal(orig_events, events)
    assert events is not orig_events


def test_transform_merge_polarities():
    orig_events, sensor_size = create_random_input()
    transform = transforms.MergePolarities()
    events = transform(orig_events)
    assert len(np.unique(orig_events["p"])) == 2
    assert len(np.unique(events["p"])) == 1
    assert events is not orig_events


def test_transform_numpy_array():
    orig_events, sensor_size = create_random_input()
    transform = transforms.NumpyAsType(int)
    events = transform(orig_events)
    assert events.dtype == int
    assert events is not orig_events


def test_transform_numpy_array_unstructured():
    orig_events, sensor_size = create_random_input()
    transform = transforms.NumpyAsType(int)
    events = transform(orig_events)
    assert events.dtype == int
    assert events is not orig_events


@pytest.mark.parametrize("delta", [10000, 5000, (2000, 5000)])
def test_transform_refractory_period(delta):
    orig_events, sensor_size = create_random_input()

    if type(delta) == tuple:
        sampled_delta = transforms.RefractoryPeriod.get_params(delta)
        assert delta[0] <= sampled_delta <= delta[1]
        assert float(sampled_delta).is_integer()
        events = transforms.functional.refractory_period_numpy(
            events=orig_events, refractory_period=sampled_delta
        )

    else:
        transform = transforms.RefractoryPeriod(delta=delta)
        events = transform(orig_events)

    assert len(events) > 0, "Not all events should be filtered"
    assert len(events) < len(
        orig_events
    ), "Result should be fewer events than original event stream"
    assert np.isin(
        events, orig_events
    ).all(), "Added additional events that were not present in original event stream"
    assert events.dtype == events.dtype
    assert events is not orig_events


@pytest.mark.parametrize(
    "variance, clip_outliers", [(30, False), (100, True), (3.5, True), (0.8, False)]
)
def test_transform_spatial_jitter(variance, clip_outliers):
    orig_events, sensor_size = create_random_input()

    transform = transforms.SpatialJitter(
        sensor_size=sensor_size,
        var_x=variance,
        var_y=variance,
        sigma_xy=0,
        clip_outliers=clip_outliers,
    )

    events = transform(orig_events)

    if not clip_outliers:
        assert len(events) == len(orig_events)
        assert (events["t"] == orig_events["t"]).all()
        assert (events["p"] == orig_events["p"]).all()
        assert (events["x"] != orig_events["x"]).any()
        assert (events["y"] != orig_events["y"]).any()
        assert np.isclose(events["x"].all(), orig_events["x"].all(), atol=2 * variance)
        assert np.isclose(events["y"].all(), orig_events["y"].all(), atol=2 * variance)

        assert (
            events["x"] - orig_events["x"]
            == (events["x"] - orig_events["x"]).astype(int)
        ).all()

        assert (
            events["y"] - orig_events["y"]
            == (events["y"] - orig_events["y"]).astype(int)
        ).all()

    else:
        assert len(events) < len(orig_events)
    assert events is not orig_events


@pytest.mark.parametrize(
    "std, clip_negative, sort_timestamps",
    [(10, True, True), (50, False, False), (0, True, False)],
)
def test_transform_time_jitter(std, clip_negative, sort_timestamps):
    orig_events, sensor_size = create_random_input()

    transform = transforms.TimeJitter(
        std=std, clip_negative=clip_negative, sort_timestamps=sort_timestamps
    )

    events = transform(orig_events)

    if clip_negative:
        assert (events["t"] >= 0).all()
    else:
        assert len(events) == len(orig_events)
    if sort_timestamps:
        np.testing.assert_array_equal(events["t"], np.sort(events["t"]))
    if not sort_timestamps and not clip_negative:
        np.testing.assert_array_equal(events["x"], orig_events["x"])
        np.testing.assert_array_equal(events["y"], orig_events["y"])
        np.testing.assert_array_equal(events["p"], orig_events["p"])
        assert (
            events["t"] - orig_events["t"]
            == (events["t"] - orig_events["t"]).astype(int)
        ).all()
    assert events is not orig_events


@pytest.mark.parametrize("p", [1, 0])
def test_transform_time_reversal(p):
    orig_events, sensor_size = create_random_input()

    original_t = orig_events["t"][0]
    max_t = np.max(orig_events["t"])

    transform = transforms.RandomTimeReversal(p=p)
    events = transform(orig_events)

    if p == 1:
        assert np.array_equal(orig_events["t"], max_t - events["t"][::-1])
        assert np.array_equal(
            orig_events["p"],
            np.invert(events["p"][::-1].astype(bool)),
        )
    elif p == 0:
        assert np.array_equal(orig_events, events)
    assert events is not orig_events


@pytest.mark.parametrize("coefficient, offset", [(3.1, 100), (0.3, 0), (2.7, 10)])
def test_transform_time_skew(coefficient, offset):
    orig_events, sensor_size = create_random_input()

    transform = transforms.TimeSkew(coefficient=coefficient, offset=offset)

    events = transform(orig_events)

    assert len(events) == len(orig_events)
    assert np.min(events["t"]) >= offset
    assert (events["t"] == (events["t"]).astype(int)).all()
    assert all((orig_events["t"] * coefficient + offset).astype(int) == events["t"])
    assert events is not orig_events


@pytest.mark.parametrize(
    "n",
    [
        100,
        0,
        (
            10,
            100,
        ),
    ],
)
def test_transform_uniform_noise(n):
    orig_events, sensor_size = create_random_input()

    if type(n) == tuple:
        sampled_n = transforms.UniformNoise.get_params(n=n)
        assert n[0] <= sampled_n <= n[1]
        assert float(sampled_n).is_integer()
        events = transforms.functional.uniform_noise_numpy(
            events=orig_events, sensor_size=sensor_size, n=sampled_n
        )
        assert len(events) == len(orig_events) + sampled_n

    else:
        transform = transforms.UniformNoise(sensor_size=sensor_size, n=n)
        events = transform(orig_events)
        assert len(events) == len(orig_events) + n

    assert np.isin(orig_events, events).all()
    assert np.isclose(
        np.sum((events["t"] - np.sort(events["t"])) ** 2), 0
    ), "Event noise should maintain temporal order."
    assert events is not orig_events


def test_transform_time_alignment():
    orig_events, sensor_size = create_random_input()

    transform = transforms.TimeAlignment()

    events = transform(orig_events)

    assert np.min(events["t"]) == 0
    assert events is not orig_events



"""
=======
ToFrame
=======

:class:`~tonic.transforms.ToFrame` creates dense frames from events by binning them in different ways.
"""

import tonic

nmnist = tonic.datasets.NMNIST("../../tutorials/data", train=False)
events, label = nmnist[0]

####################################
# ToFrame time_window
# -------------------------

transform = tonic.transforms.ToFrame(
    sensor_size=nmnist.sensor_size,
    time_window=10000,
)

frames = transform(events)

ani = tonic.utils.plot_animation(frames)


####################################
# ToFrame spike_count
# -------------------------

frame_transform = tonic.transforms.ToFrame(
    sensor_size=nmnist.sensor_size,
    event_count=100,
)

frames = transform(events)

ani = tonic.utils.plot_animation(frames)

####################################
# ToFrame n_time_bins
# ----------------------

frame_transform = tonic.transforms.ToFrame(
    sensor_size=nmnist.sensor_size,
    n_time_bins=20,
)

frames = transform(events)

ani = tonic.utils.plot_animation(frames)



import os
from typing import Callable, List, Optional, Union

import h5py
import numpy as np
from tonic.dataset import Dataset
from tonic.download_utils import (
    check_integrity,
    download_and_extract_archive,
    download_url,
    list_files,
)
from tonic.io import make_structured_array


class TUMVIE(Dataset):
    """`TUM-VIE <https://vision.in.tum.de/data/datasets/visual-inertial-event-dataset>`_

    TUM-VIE is an event camera dataset for developing 3D perception and navigation algorithms. It contains
    handheld and head-mounted sequences in indoor and outdoor environments with rapid motion during sports
    and high dynamic range. TUM-VIE includes challenging sequences where state-of-the art VIO fails or
    results in large drift. Hence, it can help to push the boundary on event-based visual-inertial algorithms.

    The dataset contains:

    * Stereo event data Prophesee Gen4 HD (1280x720 pixels)
    * Stereo grayscale frames at 20Hz (1024x1024 pixels)
    * IMU data at 200Hz
    * 6dof motion capture data at 120Hz (beginning and end of each sequence)

    Timestamps between all sensors are synchronized in hardware.

    .. note:: To be able to read this dataset, you will need `hdf5plugin` and `PIL` packages installed.

    .. note:: Use `bike-night` instead of `bike-dark` in the list of recordings if you want that specific one.

    ::

        @string{iros="International Conference on Intelligent Robots and Systems (IROS)"}
        @inproceedings{klenk2021tumvie,
         author = {S Klenk and J Chui and N Demmel and D Cremers},
         title = {TUM-VIE: The TUM Stereo Visual-Inertial Event Dataset},
         eprint = {2108.07329},
         eprinttype = {arXiv},
         eprintclass = {cs.CV},
         booktitle = {International Conference on Intelligent Robots and Systems (IROS)},
         year = {2021},
         keywords = {tumvie, event camera, dynamic vision sensor, SLAM, vslam},
        }

    Parameters:
        save_to (string): Location to save files to on disk. Will save files in a sub folder 'davis_dataset'.
        recording (string): Use the name of the recording or a list thereof to download it, for example 'skate-hard'
                            or ['skate-hard', 'bike-easy']. See project homepage for a list of available recordings.
                            Can use 'all' to download all available recordings.
        transform (callable, optional): A callable of transforms to apply to events and/or images.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
        transforms (callable, optional): A callable of transforms that is applied to both data and
                                         labels at the same time.
    """

    base_url = "https://tumevent-vi.vision.in.tum.de/"
    recordings = [
        "mocap-1d-trans",
        "mocap-3d-trans",
        "mocap-6dof",
        "mocap-desk",
        "mocap-desk2",
        "mocap-shake",
        "mocap-shake2",
        "office-maze",
        "running-easy",
        "running-hard",
        "skate-easy",
        "skate-hard",
        "loop-floor0",
        "loop-floor1",
        "loop-floor2",
        "loop-floor3",
        "floor2-dark",
        "slide",
        "bike-easy",
        "bike-hard",
        "bike-night",
    ]
    filenames = ["events_left.h5", "events_right.h5", "vi_gt_data.tar.gz"]

    sensor_size = (1280, 720, 2)
    dtype = np.dtype([("p", int), ("t", int), ("x", int), ("y", int)])
    ordering = dtype.names
    folder_name = ""

    def __init__(
        self,
        save_to: str,
        recording: Union[str, List[str]],
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        transforms: Optional[Callable] = None,
    ):
        super().__init__(
            save_to,
            transform=transform,
            target_transform=target_transform,
            transforms=transforms,
        )

        if recording == "all" or ["all"]:
            self.selection = self.recordings
        else:
            self.selection = recording if type(recording) == list else [recording]

            for recording in self.selection:
                if recording not in self.recordings:
                    raise RuntimeError(f"Recording {recording} is not available.")

        self._check_exists()

    def __getitem__(self, index):
        """
        Returns:
            tuple of (data, target), where data is a dictionary of (events_left, events_right, imu)
            and targets is a dictionary of (images_left, images_right, mocap).
        """
        base_folder = os.path.join(self.location_on_system, self.selection[index])

        import hdf5plugin  # necessary to read event files
        from PIL import Image  # necessary to read images

        events_left_file = h5py.File(
            os.path.join(base_folder, self.selection[index] + "-events_left.h5")
        )["events"]
        events_left = make_structured_array(
            events_left_file["p"][()],
            events_left_file["t"][()],
            events_left_file["x"][()],
            events_left_file["y"][()],
            dtype=self.dtype,
        )

        events_right_file = h5py.File(
            os.path.join(base_folder, self.selection[index] + "-events_right.h5")
        )["events"]
        events_right = make_structured_array(
            events_right_file["p"][()],
            events_right_file["t"][()],
            events_right_file["x"][()],
            events_right_file["y"][()],
            dtype=self.dtype,
        )

        imu_data = []
        with open(os.path.join(base_folder, "imu_data.txt")) as f:
            header = f.readline()
            for line in f.readlines():
                imu_data.append([float(num) for num in line.split()])
        imu_data = np.array(imu_data)

        mocap_data = []
        with open(os.path.join(base_folder, "mocap_data.txt")) as f:
            header = f.readline()
            for line in f.readlines():
                mocap_data.append([float(num) for num in line.split()])
        mocap_data = np.array(mocap_data)

        # images
        images_left_filenames = list_files(
            os.path.join(base_folder, "left_images"), ".jpg", prefix=True
        )
        images_left = np.stack(
            [np.array(Image.open(file)) for file in images_left_filenames]
        )
        with open(
            os.path.join(base_folder, "left_images", "image_timestamps_left.txt")
        ) as f:
            images_left_timestamps = np.array(
                [float(line) for line in f.readlines()[1:]]
            )
        images_left_timestamps -= images_left_timestamps[0]

        images_right_filenames = list_files(
            os.path.join(base_folder, "right_images"), ".jpg", prefix=True
        )
        images_right = np.stack(
            [np.array(Image.open(file)) for file in images_right_filenames]
        )
        with open(
            os.path.join(base_folder, "right_images", "image_timestamps_right.txt")
        ) as f:
            images_right_timestamps = np.array(
                [float(line) for line in f.readlines()[1:]]
            )
        images_right_timestamps -= images_right_timestamps[0]

        data = {
            "events_left": events_left,
            "events_right": events_right,
            "imu": imu_data,
        }
        targets = {
            "images_left": {"frames": images_left, "ts": images_left_timestamps},
            "images_right": {"frames": images_right, "ts": images_right_timestamps},
            "mocap": mocap_data,
        }

        if self.transform is not None:
            data = self.transform(data)
        if self.target_transform is not None:
            targets = self.target_transform(targets)
        if self.transforms is not None:
            data, targets = self.transforms(data, targets)
        return data, targets

    def __len__(self):
        return len(self.selection)

    def _check_exists(self):
        for recording in self.selection:
            file_folder = os.path.join(self.location_on_system, recording)
            os.makedirs(file_folder, exist_ok=True)
            for filename in self.filenames:
                if check_integrity(
                    os.path.join(file_folder, f"{recording}-{filename}")
                ):
                    continue
                url = f"{self.base_url}{recording}/{recording}-{filename}"
                if filename.endswith("tar.gz"):
                    download_and_extract_archive(url, file_folder)
                else:
                    download_url(url, file_folder)



try:
    import torch
except:
    ...


class PadTensors:
    """This is a custom collate function for a pytorch dataloader to load multiple
    event recordings at once. It's intended to be used in combination with sparse tensors.
    All tensor sizes are extended to the largest one in the batch, i.e. the longest recording.

    Example:
        >>> dataloader = torch.utils.data.DataLoader(dataset,
        >>>                                          batch_size=10,
        >>>                                          collate_fn=tonic.collation.PadTensors(),
        >>>                                          shuffle=True)

    """

    def __init__(self, batch_first: bool = True):
        self.batch_first = batch_first

    def __call__(self, batch):
        samples_output = []
        targets_output = []

        max_length = max([sample.shape[0] for sample, target in batch])
        for sample, target in batch:
            if not isinstance(sample, torch.Tensor):
                sample = torch.tensor(sample)
            if not isinstance(target, torch.Tensor):
                target = torch.tensor(target)
            if sample.is_sparse:
                sample.sparse_resize_(
                    (max_length, *sample.shape[1:]),
                    sample.sparse_dim(),
                    sample.dense_dim(),
                )
            else:
                sample = torch.cat(
                    (
                        sample,
                        torch.zeros(
                            max_length - sample.shape[0],
                            *sample.shape[1:],
                            device=sample.device
                        ),
                    )
                )
            samples_output.append(sample)
            targets_output.append(target)
        return (
            torch.stack(samples_output, 0 if self.batch_first else 1),
            torch.tensor(targets_output, device=target.device),
        )



from torchdata.datapipes.iter import (
    FileLister,
    Filter,
    Forker,
    Zipper,
)
from scipy.io import loadmat
import numpy as np
from pathlib import Path
from tonic.download_utils import extract_archive, check_integrity

#####
# Dataset properties.
#####

sensor_size = (10, 10, 2)
dtype = np.dtype([("x", int), ("y", int), ("t", int), ("p", int)])
MD5 = "2eef16be7356bc1a8f540bb3698c4e0ei" 

#####
# Functions
#####


def _is_mat_file(filename):
    return filename.endswith("mat") and "LUT" not in filename


def _read_label_from_filepath(filepath):
    return int(filepath.split("/")[-2])


def _at_least_n_files(root, n_files, file_type):
    check = n_files <= len(list(Path(root).glob(f"**/*{file_type}")))
    return check


def _spiketrain_to_array(matfile):
    # Transposing since the order is (address, event),
    # but we like (event, address).
    mat = loadmat(matfile)
    spiketrain = mat["spiketrain"].T
    # Separating coordinates and timestamps.
    spikes, timestamps = spiketrain[:, :-1], spiketrain[:, -1]
    # Getting events addresses.
    # First entry -> Event number.
    # Second entry -> Event address in [0,100).
    events_nums, events_addrs = spikes.nonzero()
    # Mapping addresses to 2D coordinates.
    # The mapping is (x%address, y//address), from the paper.
    events = np.zeros((len(events_nums)), dtype=dtype)
    events["x"] = events_addrs % sensor_size[0]
    events["y"] = events_addrs // sensor_size[1]
    # Converting floating point seconds to integer microseconds.
    events["t"] = (timestamps[events_nums] * 1e6).astype(int)
    # Converting -1 polarities to 0.
    events["p"] = np.max(spikes[(events_nums, events_addrs)], 0).astype(int)
    return events


def _check_exists(filepath, md5):
    check = check_integrity(filepath, md5) 
    check &= _at_least_n_files(filepath, n_files=100, file_type=".mat")
    return check


#####
# Dataset
#####


def stmnist(root, transform=None, target_transform=None):
    # The root is specified as "directory/archive.zip". 
    # We strip 'archive.zip' and get only 'directory', where it is
    # extracted in 'data_submission'. 
    filepath = root[::-1].split("/", 1)[-1][::-1] + "/data_submission"
    # Extracting the ZIP archive.
    if not _check_exists(filepath, MD5):
        extract_archive(root)
    # Creating the datapipe.
    dp = FileLister(root=filepath, recursive=True)
    dp = Filter(dp, _is_mat_file)
    event_dp, label_dp = Forker(dp, num_instances=2)
    event_dp = event_dp.map(_spiketrain_to_array)
    label_dp = label_dp.map(_read_label_from_filepath)
    if transform:
        event_dp = event_dp.map(transform)
    if target_transform:
        label_dp = label_dp.map(target_transform)
    dp = Zipper(event_dp, label_dp)
    return dp



import numpy as np
from tonic.slicers import (
    SliceAtIndices,
    SliceAtTimePoints,
    SliceByEventCount,
    SliceByTime,
    slice_events_at_indices,
)


def test_slice_at_indices_class():

    slicer = SliceAtIndices(start_indices=[0, 3, 5], end_indices=[3, 5, 7])
    data = np.arange(7)

    slices, _ = slicer.slice(data, None)
    assert len(slices) == 3


def test_slice_at_indices_method():

    data = np.arange(7)
    slices = slice_events_at_indices(
        data, start_indices=[0, 3, 5], end_indices=[3, 5, 7]
    )

    assert len(slices) == 3


def test_slice_at_time_points_class():
    data = np.arange(100).astype(dtype=[("t", float)])
    slicer = SliceAtTimePoints(start_tw=[10, 20, 50], end_tw=[20, 50, 100])

    slices, _ = slicer.slice(data=data, targets=None)

    assert len(slices) == 3


def test_slice_by_event_count():
    data = np.arange(100)
    slicer = SliceByEventCount(event_count=50, overlap=25, include_incomplete=False)

    data_slice, _ = slicer.slice(data, None)

    assert len(data_slice) == 3


def test_slice_by_time():
    data = np.arange(101).astype(dtype=[("t", float)])
    slicer = SliceByTime(time_window=5)

    slices, _ = slicer.slice(data, None)

    assert len(slices) == 20
